/************************************************************************
NAME: flecs90
DESCRITPTION: Translates source code from FLECS (Fortran Language
              Extended Control Structures) to FORTRAN 90 code.
PARAMETER: name - name of FLECS source code file
DATE: August 2, 1994
*************************************************************************/

/* INCLUDE FILES */
#include "flecs90.h"

/* FUNCTION DEFINITION */
void flecs90(char *name)

{
/* VARIABLES */

  /* External */
  extern char *prog_name;   /* name of program */
  extern int qfortran;      /* if true a fortran output file is
                               produced */
  extern FILE *flecs;       /* file pointer to flecs input file */
  extern FILE *fortran;     /* file pointer to fortran output file */
  extern char *flecs_name;  /* name of file with flecs source code */
  extern char *fortran_name;/* name of file with fortran code */
  extern int qcont;         /* TRUE if main processing loop should continue */
  extern int qto;           /* TRUE if flecs in-line procedures have
			       started */
  extern int nifthen;       /* current number of active fortran IF-THEN
			       statements */
  extern int nloop;         /* Number of flecs loops encountered */
  extern int line_num;      /* current line number in flecs source file */
  extern char *line;        /* current line from flecs source file */
  extern int line_len;      /* length of the current line */
  extern enum state_type state; 
                            /* state of code being processed */
  extern enum key_type key;   /* current flecs key word */
  extern struct key_stacks *fin_stack;
                            /* stack of flecs commands which are not
			       finished yet */
  extern struct key_stacks *cycle_stack;
                            /* stack of command which can have CYCLE
			       or EXIT called */
  extern struct line_stacks *repeat_stack;
                            /* stack of instances associated with
			       REPEAT WHILE or REPEAT UNTIL statememnt */
  extern struct line_stacks *select_stack;
                            /* stack of ref_spec associated with
			       SELECT or CASE statement */
  extern int errnum;        /* Total number of flecs errors in a
			       subprogram */
  extern char *error_msg;   /* error message string */
  extern char *flecs_proc;  /* name of in-line flecs procedure calls */
  extern char *scratch;     /* length 80 string for temporary use */
  extern char *keyword[NKEYS+1];
                            /* Vector of Flecs Keywords as Characters */

  /* Local */
  char *extension;          /* extension on input file */
  int base_length;          /* length of base file name without extension */
  int cur_col;              /* current column in the flecs source file */
  int i;                    /* loop control variables */

/* BEGIN CODE */
/************************************************************************
Initalize Variables:
*************************************************************************/
  flecs_proc = (char *) calloc(80,sizeof(char));
  error_msg = (char *) calloc(81,sizeof(char));
  scratch = (char *) calloc(80,sizeof(char));
  line = (char *) calloc(80,sizeof(char));
  if ((flecs_proc == NULL) || (line == NULL) || (error_msg == NULL)
      || (scratch == NULL)) {
    fprintf(stderr,"%s\n","memory error");
    exit(1);
  }
  key = NONE;
  state = STANDARD;
  fin_stack = NULL;
  cycle_stack = NULL;
  repeat_stack = NULL;
  select_stack = NULL;
  line_num = 0;
  qcont = 1;
  qto = 0;
  nifthen = 0;
  nloop = 0;
  errnum = 0;

/************************************************************************
Open flecs source file and fortran code file:
*************************************************************************/
  /* locate the extension and the length of the pre-extension (base)
     file name, <filename> */
  extension = name;
  base_length = 0;
  while ((*extension != '.') && (*extension != '\0')) {
    extension++;
    base_length++;
  }

  /* issue and error message if a file with the wrong extension
     is used */
  if ((*extension != '\0') && (!eq(extension,".flx"))) {
    fprintf(stderr,"%s: %s contains an incorrect file suffix\n",
	    prog_name,name);
    quit();
    return;
  }

  /* define enough memory for the flecs and fortran file names 
     including extensions */
  flecs_name = (char *) malloc((base_length+5)*sizeof(char));
  fortran_name = (char *) malloc((base_length+3)*sizeof(char));
  if ((flecs_name == NULL) || (fortran_name == NULL)) {
    fprintf(stderr,"%s\n","memory error");
    exit(1);
  }

  /* Copy the base file name to the flecs and fortran file names */
  for (i=0; i<base_length; i++) {
    *(flecs_name+i) = *(name+i);
    *(fortran_name+i) = *(name+i);
  }

  /* Copy the appropriate extensions to the flecs and fortran file
     names */
  *(flecs_name+base_length) = '.';
  *(flecs_name+base_length+1) = 'f';
  *(flecs_name+base_length+2) = 'l';
  *(flecs_name+base_length+3) = 'x';
  *(flecs_name+base_length+4) = '\0';

  *(fortran_name+base_length) = '.';
  *(fortran_name+base_length+1) = 'f';
  *(fortran_name+base_length+2) = '\0';

  /* Open the flecs file for reading */
  if ((flecs = fopen(flecs_name,"r")) == NULL) {
    fprintf(stdout,"%s: %s not found\n",prog_name,flecs_name);
    quit();
    return;
  }

  /* Open the fortran file for writting. Note that it will be
     overwritten if it already exists */
  if (qfortran) fortran = fopen(fortran_name,"w");
/************************************************************************
GET THE FIRST LINE OF THE FLECS SOURCE FILE:
*************************************************************************/
  if (!getline()) {
    fprintf(stderr,"%s: %s file empty\n",prog_name,flecs_name);
    quit();
    return;
  }
/************************************************************************
LOOP THROUGH EACH LINE OF THE FLECS SOURCE FILE AND PARSE IT
*************************************************************************/
  while (qcont) {
    
    /* get key word from line */
    key = getkey(line,line_len,&cur_col);

/************************************************************************
IF END WAS ENCOUNTERED TERMINATE A SUBPROGRAM:
*************************************************************************/
    if (key == END) {

      /* Check the stack for unfinished structures */
      while (fin_stack != NULL) {
	sprintf(error_msg,"%s (line %d) not finished before END.",
		keyword[fin_stack->key],fin_stack->line_num);
	flecs_error(1,error_msg,line_num);
	if (fin_stack->key == WHEN) fin_stack->key = IF;
	finish();
      }

      /* Echo the end line */
      if (qfortran) {
	for (i=1; i <= line_len; i++) fputc(line[i],fortran);
	print_line_number(line_len,line_num);
      }

      /* Get the next line. If it exists continue otherwise quit */
      if (!(qcont = getline())) {
	quit();
	return;
      }

      /* Reinitialize variables */
      key = NONE;
      state = STANDARD;
      fin_stack = NULL;
      cycle_stack = NULL;
      repeat_stack = NULL;
      select_stack = NULL;
      qcont = 1;
      qto = 0;
      nifthen = 0;
      nloop = 0;
      errnum = 0;

    } else {
/************************************************************************
DEAL WITH KEYWORD BASED ON THE CURRENT STATE
*************************************************************************/
      switch (state) {
	
      /*******************************************************/
      /* Parse a command that is not in the standard state.  */
      /* For these states certain input is required.         */ 
      /*******************************************************/
      case BRANCH_FIN: case OTHERWISE_FIN: case WHEN_FIN: case TO_FIN:
	if ((state == TO_FIN) && (key != TO)) {
	  flecs_error(0,"Only TO or END valid.",line_num);
	  qcont = getline();
	  break;
	} else if ((state == WHEN_FIN) && (key != ELSE)) {
	  sprintf(error_msg,"Only ELSE valid. (WHEN line %d)",
		  fin_stack->line_num);
	  flecs_error(1,error_msg,line_num);
	  fin_stack->key = IF;
	  finish();
	} else if (((state == BRANCH_FIN) && (key != FIN) && (key != INSTANCE))
		   || ((state == OTHERWISE_FIN) && (key != FIN))) {
	  if ((state == OTHERWISE_FIN) && (key == INSTANCE)) {
	    sprintf(error_msg,"OTHERWISE completed. Only FIN valid."
		    " (%s line %d)",keyword[fin_stack->key],
		    fin_stack->line_num);
	    flecs_error(1,error_msg,line_num);
	  } else {
	    sprintf(error_msg,"Only INSTANCE or FIN valid. (%s line %d)",
		    keyword[fin_stack->key],fin_stack->line_num);
	    flecs_error(1,error_msg,line_num);
	  }
	  finish();
	}

      /******************************************/
      /* Parse a command in the standard state  */
      /* (any input is acceptable)              */
      /******************************************/
      case STANDARD:
	state = STANDARD;
	process_standard(cur_col);
      }
    }
  }
/************************************************************************
EOF without END statement
*************************************************************************/
  /* Check the stack for unfinished structures */
  while (fin_stack != NULL) {
    sprintf(error_msg,"%s (line %d) not finished before end of file.",
	    keyword[fin_stack->key],fin_stack->line_num);
    flecs_error(1,error_msg,line_num);
    if (fin_stack->key == WHEN) fin_stack->key = IF;
    finish();
  }
  fprintf(stderr,"\"%s\", line %d:     %s\n",flecs_name,line_num,
	  "Missing END statement.");
  quit();
  return;
}
