/************************************************************************
NAME: flecs90_util.c
DESCRITPTION: Utilities for the flecs90 program.
UTILITES: 
          eq - tests two strings for equality.
	  length - returns the length of a string
          getline - reads in the next line of the flecs source file
          print_line_number - prints a line number to the fortran
	                      file
	  echo_label - print the first 5 lines of a flecs source
	               file in the fortran code file
	  echo_label_continue - print a label if one exists with
	                        a fortran continue statement
	  getword - gets the next word from a line of the flecs 
	            source file
	  qproc - Finds a flecs procedure name from a line of text
	  getinstance - gets the next instance "(anything)" from
	                a flecs command
	  key_match - checks for a match with a keyword
	  getkey - gets the flecs keyword from the current line
	  qcontinue - checks to see if a line is a continuation line
	  add_to_command - adds a line onto the current flecs command
	  fortran_put - writes part of an arry to the fortran
	                file and colapses all white space into
			1 space
	  parse_cont - parses a continued flecs command
	  finish - finishes a flecs structure.
	  push_key - enters a key_type and line number at the top 
	             of a key stack
	  pop_key - removes the top structure from a key stack
	  push_line - enters a character string at the top of a line stack
	  pop_line - removes the top structure from a line stack
	  flecs_error - prints out a flecs error message
	  quit - frees memory and closes files for flecs90 function
DATE: August 3, 1994
*************************************************************************/
/* INCLUDE FILES */
#include "flecs90.h"
#include <ctype.h>

/************************************************************************
EQ:
tests two strings for equality
************************************************************************/
int eq(char *a,char *b)
{
   while ( (*a == *b) && (*a != '\0') && (*b != '\0')) {
      a++;
      b++;
   }
   return((*a == *b));
}
/************************************************************************
LENGTH:
returns the length of a string
************************************************************************/
int length(char *a)
{
  char *b = a;

  while (*b != '\0') b++;
  return(b-a);
}

/************************************************************************
GETLINE:
-reads in the next line of the flecs source file
-If it is a comment GETLINE echos the comment to the fortran file
 and gets the next line
-If it is blank GETLINE increments the blank line counter and
 gets the next line
-return TRUE if a line was found; FALSE otherwise
-Also sets qcont to FALSE if no line was found
************************************************************************/
int getline(void)
{

  /* External Variables */
  extern int qcomment;
  extern FILE *flecs;
  extern FILE *fortran;
  extern char *line;
  extern int line_num;
  extern int line_len;
  extern int qcont;

  /* Local Variables */
  int c;                 /* current character */
  int i,j;               /* loop control variables */
  int space;             /* number of spaces */
  int qeof=0;            /* True if an EOF was found on the line */
  int qblank;            /* True if the current line is blank */
  int num_prev_blnk=0;   /* Number of blank lines found */
  
  while(1) {

    /* Initalize variables */
    if (++line_num > 99999) line_num = 1;
    line_len = 0;
    space = 0;
    qblank=1;
    
    /* Read in Line */
    for (i=1; i<=FORT_LINE; i++) {
      c = fgetc(flecs);
      if (c == EOF) {
	qeof = 1;
	break;
      } else if (c == '\n') break;
      else  {
	if (isspace(c)) space++;
	else {
	  for (j=1; j<=space; j++) line[line_len+j] = ' ';
	  line_len += space+1;
	  line[line_len] = c;
	  qblank = 0;
	  space = 0;
	}
      }
    }
    
    /* If the line continues past column 72 read it in but ignore it */
    if (i > FORT_LINE)
      while (((c = fgetc(flecs)) != EOF) && (c != '\n')) ;
    
    /* If the line is blank and end of file was found return false */
    if (qeof && qblank) return((qcont = 0));
    
    /* If the line is blank increment the blank line counter */
    if (qblank) num_prev_blnk++;
    
    /* Otherwise check for a comment */
    else {
      
      /* Echo previous blank lines if neccessary */
      if (qcomment)
	for (i=1; i <= num_prev_blnk; i++) fputc('\n',fortran);
      num_prev_blnk = 0;

      /* If the line is a comment and comments are being echoed to
	 the fortran file then output the line */
      if ((toupper(line[1]) == 'C') || (line[1] == '*')) {
	if (qcomment) {
	  for (i=1; i <= line_len; i++) fputc(line[i],fortran);
	  print_line_number(line_len,line_num);
	}
      }
      
      /* Otherwise it is not a comment so we have found a line */
      else return((qcont = 1));
    }
  }
}

/************************************************************************
PRINT_LINE_NUMBER:
prints a line number to the fortran file
Paramters
  (1) start -> column number line is currently at
  (2) num -> line number to be printed
************************************************************************/
void print_line_number(int start, int num)
{
  /* External Variables */
  extern int qline;
  extern FILE *fortran;
  
  /* Local Variables */
  int i;   /* loop control variable */
  
  if (qline) {
    for (i=(start+1);i<=(FORT_LINE+1);i++) fputc(' ',fortran);
    if (num <= 9) fprintf(fortran,"0000%d",num);
    else if (num <= 99) fprintf(fortran,"000%d",num);
    else if (num <= 999) fprintf(fortran,"00%d",num);
    else if (num <= 9999) fprintf(fortran,"0%d",num);
    else fprintf(fortran,"%d",num);
  }
  fputc('\n',fortran);
}

/************************************************************************
ECHO_LABEL:
print the first 5 lines of a flecs source file in the fortran code file
Paramters
  (1) lline -> line
************************************************************************/
void echo_label(char *lline)
{
  /* External Variables */
  extern FILE *fortran;
  
  /* Local Variables */
  int i;   /* loop control variable */
  
  for (i=1; i <= 5; i++) fputc(lline[i],fortran);
  fputc(' ',fortran);
}

/************************************************************************
ECHO_LABEL_CONTINUE:
print a label if one exists with a fortran continue statement
Paramters
  (1) lline -> line
  (2) lline_num -> current line number
************************************************************************/
void echo_label_continue(char *lline,int lline_num)
{
  /* External Variables */
  extern int qline;
  extern FILE *fortran;
  
  /* Local Variables */
  int i;        /* loop control variable */
  int fort_loc; /* column location in the fortran output file */
  int nblank;   /* TRUE if the label contains non-blank characters */
  
  nblank = 0;
  for (i=1; i <= 5; i++)
    if (!isspace(lline[i])) nblank = 1;
  if (nblank) {
    echo_label(lline);
    fort_loc = 6;
    fortran_put("CONTINUE",0,7,&fort_loc,lline_num);
    print_line_number(fort_loc,lline_num);
  }
}

/************************************************************************
GETWORD:
gets the first word beginning in column "first"
from a line of the flecs source file
returns TRUE if a word was found; FALSE otherwise
Paramters
  (1) lline -> the line
  (2) len -> length of the line (location of last non-blank character)
  (3) begin -> the column to begin the search
  (4) start <- the column in which the word begins
  (5) end <- the column in which the word ends
************************************************************************/
int getword(char *lline,int len,int begin,int *start,int *end)
{
  int i;           /* loop control variables */
  int qspace;      /* TRUE is the current character is blank space */
  
  /* Initalize variables */
  *start = 0;
  *end = 0;
  
  /* Find word */
  for (i=begin; i<=len; i++) {
    qspace = isspace(lline[i]);
    /* If this is the first non-white space start the word */
    if ((*start == 0) && (!qspace)) {
      *start = i;
    }
    
    /* If this is space at the end of the word; end the word */
    else if ((*start != 0) && (*end == 0) && (qspace || (lline[i] == '('))) {
      *end = i-1;
      break;
    }
  }
  
  /* If start != 0 and end == 0, the word terminated at the end
     of the line */
  if ((*start != 0) && (*end == 0)) *end = len;
  
  /* Return start which will be 0 or false if we didn't find a word */
  return(*start);
}

/************************************************************************
QPROC:
Finds a flecs procedure name from a line of text
returns TRUE if one was found; false otherwise
Paramters
  (1) lline -> the line
  (2) start -> the column in which the word begins
  (3) end -> the column in which the word ends
************************************************************************/
int qproc(char *lline,int start,int end)
{
  /* External Variables */
  extern char *flecs_proc;

  /* Local Variables */
  int i;            /* Loop control variable */
  int qminus=0;     /* TRUE if a minus sign is found */
  char *new;        /* pointer to flecs_proc */

  new = flecs_proc;
  if (!isalnum(lline[start])) return(0);
  else *new++ = toupper(lline[start]);
  
  for (i=(start+1); i <= end; i++) {
    if (lline[i] == '-') {
      qminus = 1;
      *new++ = '_'; 
    }
    else if (!isalnum(lline[i])) return(0);
    else *new++ = toupper(lline[i]);
  }
  *new = '\0';
  return(qminus);
}

/************************************************************************
GETINSTANCE:
gets the next instance "(anything)" from a flecs command
returns 
  0 : no instance was found
  1 : a complete instance was found
  2 : an incomplete instance was found
Paramters
  (1) lline -> the line
  (2) len -> length of the line (location of last non-blank character)
  (3) begin -> the column to begin the search
  (4) start <- the column in which the instance begins
  (5) end <- the column in which the instance ends
************************************************************************/
int getinstance(char *lline,int len,int begin,int *start,int *end)
{
  int i;         /* loop control variables */
  int paren;       /* total number of open parenthesis */

  /* Initalize variables */
  *start = 0;
  *end = 0;
  
  /* Find word */
  for (i=begin; i<=len; i++) {

    if (!isspace(lline[i])) {

      /* If this is the first non-white space start the instance */
      if (*start == 0) {
	if (lline[i] != '(') return(0);
	paren = 1;
	*start = i;
      }
    
      /* Otherwise Check to see if it is an open parenthesis */
      else if (lline[i] == '(') paren++;

      /* Check to see if it is a closing parenthesis;
	 if so see if it closes the instance */
      else if (lline[i] == ')') {
	paren--;
	if (paren == 0) {
	  *end = i;
	  return(1);
	}
      }
    }
  }
  
  /* If we get this far either the instance was never started
     or it was not completed */
  if (*start == 0) return(0);
  else return(2);
}

/************************************************************************
KEY_MATCH:
Checks for a match between a keyword and a portion of a flecs
source code line
returns TRUE if there is a match
Parameters
  (1) lline -> the line
  (2) begin -> the beginning of the word
  (3) end -> the end of the word
  (4) lkey -> a character string representing the keyword
************************************************************************/
int key_match(char *lline,int start,int end,char *lkey)
{
  int i;      /* loop control variable */
  int qans=1; /* TRUE if key and word match */

  if ((end - start + 1) != ((int)length(lkey))) return(0);
  for (i=start; i <= end; i++)
    if (toupper(lline[i]) != *lkey++) {
      qans = 0;
      break; 
    }

  return (qans);
}

/************************************************************************
GETKEY:
gets the flects keyword from the current line
returns key of type enum key_type indicating the flecs key word
found on that line
Paramters
  (1) lline -> the line
  (2) lline_len -> length of the line (location of last non-blank character)
  (3) cur_col <- the column which ends the flecs keyword
************************************************************************/
enum key_type getkey(char *lline,int lline_len,int *cur_col)
{
  /* External Variables */
  extern char *keyword[NKEYS+1];

  /* Local Variables */
  int i;              /* loop control variable */
  int start;          /* start of first word on line*/
  int end;            /* end of first word on line*/
  int start2;         /* start of second word on line */
  int end2;           /* end of second word on line */
  int qword;          /* true if there is a word in the line */
  enum key_type lkey; /* flecs key found */
  
  lkey = NONE;

  /* If it is a continuation line return with key NONE */
  if ((lline[6] != ' ') && (lline[6] != '0')) return(lkey);

  qword = getword(lline,lline_len,7,&start,&end);

  if (qword) {
    *cur_col = end;
    /* Check to see if the word is a flecs procedure */
    if (qproc(lline,start,end)) return(PROCEDURE_CALL);
    /* If the word begins with ( we have an instance from a flecs
       case, conditional or select statement */
    else if (lline[start] == '(') lkey = INSTANCE;
    /* Check for the word REPEAT. If we find it check for the
       next word being WHILE or UNTIL */
    else if (key_match(lline,start,end,"REPEAT")) {
      if (getword(lline,lline_len,end+1,&start2,&end2)) {
	if (key_match(lline,start2,end2,"UNTIL")) {
	  *cur_col = end2;
	  lkey = REPEAT_UNTIL;
	} else if (key_match(lline,start2,end2,"WHILE")) {
	  *cur_col = end2;
	  lkey = REPEAT_WHILE;
	}
      }
    }
    /* Check for the word DO if the next non-space begins with "("
       set key = DO otherwise key = NONE; */
    else if (key_match(lline,start,end,"DO")) {
      if (getword(lline,lline_len,end+1,&start2,&end2))
	if (lline[start2] == '(') lkey = DO;
    }
    /* Check for the word CASE if there is another word after case
       it is a CASE_SELECT otherwise it is a CASE_CONDITONAL */
    else if (key_match(lline,start,end,"CASE")) {
      if (lline_len > end) lkey = CASE_SELECT;
      else lkey = CASE_CONDITIONAL;
    }
    /* Check for the word being END */
    else if (key_match(lline,start,end,"END")) {
      if (getword(lline,lline_len,end+1,&start2,&end2)) {
	if (key_match(lline,start2,end2,"IF")) {
	  *cur_col = end2;
	  lkey = ENDIF;
	}
      } else lkey = END;
    }
    /* Check for any other flecs key word */
    else {
      for (i=ENDIF; i <= REVERT; i++)
	if (key_match(lline,start,end,keyword[i])) lkey = i;
    }
  }
  return(lkey);
}

/************************************************************************
QCONTINUE
checks to see if a line is continuation
return TRUE if it is a continuation line; false otherwise
Parameters
  (1) lline -> current line
************************************************************************/
int qcontinue(char *lline)
{
  return((lline[6] != ' ') && (lline[6] != '0'));
}

/************************************************************************
ADD_TO_COMMAND
adds a line onto the current flecs command
Parameters
  (1) command <-> updated with new line
  (2) add -> integer array of characters to be added to command
  (4) start -> column to begin in
  (5) end -> column to end in
  (4) command_len <-> length of the current command
************************************************************************/
char *add_to_command(char *command,char *add,int start,int end,
		     int *command_len)
{
  int i;     /* loop control variable */
  char *new; /* pointer to new command */

  new = (char *) malloc(((*command_len)+(end-start)+5)*sizeof(char));
  if (new == NULL) {
    fprintf(stderr,"%s\n","memory error");
    exit(1);
  }

  if ((*command_len) >= 1) {
    for (i=1; i <= (*command_len); i++) new[i] = command[i];
    new[++(*command_len)] = ' ';
  }

  for (i=start; i <= end; i++)
    new[++(*command_len)] = add[i];

  free(command);
  return(new);
}

/************************************************************************
FORTRAN_PUT:
writes part of an array to the fortran
file and colapses all white space into 1 space
Paramters
  (1) lline -> array from which characters will be added
  (2) start -> begin location of characters to be written
  (3) end -> end location of characters to be written
  (4) fort_loc <-> current column in the fortran output file
  (5) lline_num -> current line number in flecs source code
************************************************************************/
void fortran_put(char *lline,int start,int end,int *fort_loc,int lline_num)
{
  /* External Variables */
  extern int qline;
  extern FILE *fortran;

  /* Local Variables */
  int i;             /* loop control variable */
  int qprev_space=0; /* TRUE if the previous character is a space */

  for (i=start; i <= end; i++) {
    if (isspace(lline[i])) qprev_space = 1;
    else {
      if (qprev_space) {
	if (*fort_loc >= (FORT_LINE-1)) {
	  print_line_number(*fort_loc,lline_num);
	  fprintf(fortran,"     +");
	  *fort_loc = 6;
	} else {
	  fputc(' ',fortran);
	  (*fort_loc)++;
	}
      }
      if (*fort_loc >= FORT_LINE) {
	print_line_number(*fort_loc,lline_num);
	fprintf(fortran,"     +");
	*fort_loc = 6;
      }
      fputc(lline[i],fortran);
      (*fort_loc)++;
      qprev_space = 0;
    }
  }
  if (qprev_space) {
    if (*fort_loc >= (FORT_LINE-1)) {
      print_line_number(*fort_loc,lline_num);
      fprintf(fortran,"%s","     +");
      *fort_loc = 6;
    } else {
      fputc(' ',fortran);
      (*fort_loc)++;
    }
  }
}


/************************************************************************
PARSE_CONT
parses a continued flecs command. The following keys are recognized:
PROCEDURE_CALL, CYCLE, EXIT and REVERT.
returns TRUE if the command was continued; FALSE otherwise
Paramters
  (1) flecs_in -> character string containing flecs command
  (2) flecs_len -> length of command
  (3) word_start -> begining of word
  (4) word_end -> ending of word
  (5) fort_loc -> current column in the fortran output file
  (6) lline_num -> current line number in flecs source code
************************************************************************/
void parse_cont(char *flecs_in,int flecs_len,int word_start,int word_end,
		int fort_loc,int lline_num)
{
  /* External Variables */
  extern int qline;
  extern int qlabel;
  extern FILE *fortran;
  extern int qfortran;
  extern char *flecs_proc;
  extern char *scratch;
  extern char *error_msg;
  extern struct key_stacks *cycle_stack;
  extern int qto;
  extern char *keyword[NKEYS+1];

  /* Local Variables */
  int len;           /* length of flecs procedure name */
  int flecs_loc;     /* current location in flecs command */
  enum key_type key2;/* possible key in continued command */

  if (qproc(flecs_in,word_start,word_end)) {
    len = length(flecs_proc)-1;
    if (qfortran) {
      fortran_put("CALL ",0,4,&fort_loc,lline_num);
      fortran_put(flecs_proc,0,len,&fort_loc,lline_num);
      print_line_number(fort_loc,lline_num);
    }
  } else {
    flecs_loc = word_start;
    key2 = NONE;
    if (word_end >= flecs_len) {
      if (key_match(flecs_in,word_start,word_end,"CYCLE")) 
	key2 = CYCLE;
      else if (key_match(flecs_in,word_start,word_end,"REVERT"))
	key2 = REVERT;
      else if (key_match(flecs_in,word_start,word_end,"EXIT"))
	key2 = EXIT;
    }
    switch (key2) {
    case CYCLE: case EXIT:
      /* If there is no looping structure on the stack give an error */
      if (cycle_stack == NULL) {
	sprintf(error_msg,"Nothing to %s\n",keyword[key2]);
	flecs_error(0,error_msg,lline_num);
	if (qfortran) {
	  fortran_put("CONTINUE",0,7,&fort_loc,lline_num);
	  print_line_number(fort_loc,lline_num);
	}
      }
      /* Othewise print CYCLE or EXIT and structure name */
      else {
	if (qfortran) {
	  if (key2 == CYCLE)
	    fortran_put("CYCLE",0,4,&fort_loc,lline_num);
	  else 
	    fortran_put("EXIT",0,3,&fort_loc,lline_num);
	  sprintf(scratch,"FLECS_LOOP_%d",cycle_stack->line_num);
	  len = length(scratch) - 1;
	  if (qlabel) {
	    fortran_put(" ",0,0,&fort_loc,lline_num);
	    fortran_put(scratch,0,len,&fort_loc,lline_num);
	  }
	  print_line_number(fort_loc,lline_num);
	}
      }
      break;
    case REVERT:
      /* If we are instide a TO statement print RETURN to fortran file */
      if (qto) {
	if (qfortran) {
	  fortran_put("RETURN",0,5,&fort_loc,lline_num);
	  print_line_number(fort_loc,lline_num);
	}
      }
      /* Otherwise give an error message */
      else {
	flecs_error(0,"Nothing to REVERT",lline_num);
	if (qfortran) {
	  fortran_put("CONTINUE",0,7,&fort_loc,lline_num);
	  print_line_number(fort_loc,lline_num);
	}
      }
      break;
    default:
      if (qfortran) {
	fortran_put(flecs_in,word_start,flecs_len,&fort_loc,lline_num);
	print_line_number(fort_loc,lline_num);
      }
    }
  }
}

/************************************************************************
FINISH
Finishes the flecs structure on the top of the fin stack.
************************************************************************/
void finish(void)
{
  /* External Variables */
  extern int qfortran;      /* if true a fortran output file is
                               produced */
  extern int qselect;       /* if true SELECT (ref_spec) and CASE (ref_spec)
                               are translated to fortran 90 CASE
                               SELECT statements */
  extern int line_num;      /* current line number in flecs source file */
  extern char *line;        /* current line from flecs source file */
  extern enum state_type state; 
                            /* state of code being processed */
  extern struct key_stacks *fin_stack;
                            /* stack of flecs commands which are not
			       finished yet */
  extern struct key_stacks *cycle_stack;
                            /* stack of command which can have CYCLE
			       or EXIT called */
  extern struct line_stacks *repeat_stack;
                            /* stack of instances associated with
			       REPEAT WHILE or REPEAT UNTIL statememnt */
  extern struct line_stacks *select_stack;
                            /* stack of ref_spec associated with
			       SELECT or CASE statement */
  extern char *scratch;     /* length 80 string for temporary use */

  /* Local Variables */
  int fort_loc;             /* Current column in the fortran file */
  int len;                  /* Length of strings */

  /* BEGIN CODE */
  /* Finish the structure based on the keyword */
  switch (fin_stack->key) {
	  	  
  /* If TO is finished enter TO_FIN state; the next
     statement must be TO or END */
  case TO:
    state = TO_FIN;
    /* print END SUBROUTINE */
    if (qfortran) {
      echo_label(line);
      fort_loc = 6;
      fortran_put("END SUBROUTINE",0,13,&fort_loc,line_num);
      print_line_number(fort_loc,line_num);
    }
    break;

  /* The following cases require a change in state when
     finished but no output */
  case WHEN: case INSTANCE: case OTHERWISE:
    if (fin_stack->key == WHEN) state = WHEN_FIN;
    else if (fin_stack->key == INSTANCE) state = BRANCH_FIN;
    else if (fin_stack->key == OTHERWISE) state = OTHERWISE_FIN;
    /* if there is no label print nothing otherwise print
       the label with a continue statement */
    if (qfortran) echo_label_continue(line,line_num);
    break;

  /* IF qselect is true CASE and CASE_SELECT are terminated with
     an END SELECT statement */
  case SELECT: case CASE_SELECT:
    if (qselect) {
      if (qfortran) {
	echo_label(line);
	fort_loc = 6;
	fortran_put("END SELECT",0,9,&fort_loc,line_num);
	print_line_number(fort_loc,line_num);
      }
      break;
    }

  /* The following keys are terminated with an END IF statement */
  case IF: case ELSE: case UNLESS: case CONDITIONAL: 
  case CASE_CONDITIONAL:
    if (qfortran) {
      echo_label(line);
      fort_loc = 6;
      fortran_put("END IF",0,5,&fort_loc,line_num);
      print_line_number(fort_loc,line_num);
      if ((fin_stack->key == SELECT) || (fin_stack->key == CASE_SELECT))
	select_stack = pop_line(select_stack);
    }
    break;

  /* The following keys are one pass loops which require a
     IF (condition) EXIT statement and an ENDDO statement */
  /* They must be removed from the cycle_stack */
  case REPEAT_WHILE: case REPEAT_UNTIL:
    if (qfortran) {
      echo_label(line);
      fort_loc = 6;
      fortran_put("IF ",0,2,&fort_loc,line_num);
      if (fin_stack->key == REPEAT_WHILE)
	fortran_put("(.NOT.",0,5,&fort_loc,line_num);
      len = length(repeat_stack->line) - 1;
      fortran_put(repeat_stack->line,0,len,&fort_loc,line_num);
      if (fin_stack->key == REPEAT_WHILE)
	fortran_put(")",0,0,&fort_loc,line_num);
      fortran_put(" EXIT",0,4,&fort_loc,line_num);
      sprintf(scratch,"FLECS_LOOP_%d",cycle_stack->line_num);
      len = length(scratch) - 1;
      if (qlabel) {
	fortran_put(" ",0,0,&fort_loc,line_num);
	fortran_put(scratch,0,len,&fort_loc,line_num);
      }
      print_line_number(fort_loc,line_num);
      fprintf(fortran,"      ");
      fort_loc = 6;
      fortran_put("END DO",0,5,&fort_loc,line_num);
      if (qlabel) {
	fortran_put(" ",0,0,&fort_loc,line_num);
	fortran_put(scratch,0,len,&fort_loc,line_num);
      }
      print_line_number(fort_loc,line_num);
    }
    repeat_stack = pop_line(repeat_stack);
    cycle_stack = pop_key(cycle_stack);
    break;
	  
  /* The following keys are terminated with an ENDDO statement */
  /* They must be removed from the cycle_stack */
  case DO: case FOR: case WHILE: case UNTIL: case LOOP:
    if (qfortran) {
      echo_label(line);
      fort_loc = 6;
      fortran_put("END DO",0,5,&fort_loc,line_num);
      sprintf(scratch,"FLECS_LOOP_%d",cycle_stack->line_num);
      len = length(scratch) - 1;
      if (qlabel) {
	fortran_put(" ",0,0,&fort_loc,line_num);
	fortran_put(scratch,0,len,&fort_loc,line_num);
      }
      print_line_number(fort_loc,line_num);
    }
    cycle_stack = pop_key(cycle_stack);
    break;
  }
  if (fin_stack->key != WHEN) fin_stack = pop_key(fin_stack);
}

/************************************************************************
PUSH_KEY
Enters a key_type and line number at the top of a key stack
Returns a pointer to the new stack
Paramters
  (1) stack -> current stack
  (2) key -> key_type to be added in new structure
  (3) lline_num -> line number to be added to new structure
************************************************************************/
struct key_stacks *push_key(struct key_stacks *stack,
                           enum key_type lkey,int lline_num)
{
  struct key_stacks *root;

  root = (struct key_stacks *) malloc(sizeof(struct key_stacks));
  if (root == NULL) {
    fprintf(stderr,"memory error\n");
    exit(1);
  }
  
  root->key = lkey;
  root->line_num = lline_num;
  root->next = stack;
  return(root);
}

/************************************************************************
POP_KEY
Removes the top structure from a key stack
Returns a pointer to the new stack
Paramters
  (1) stack -> current stack
************************************************************************/
struct key_stacks *pop_key(struct key_stacks *stack)
{
  struct key_stacks *root;

  if (stack == NULL) return(NULL);
  root = stack->next;
  free(stack);
  return(root);
}

/************************************************************************
PUSH_LINE
Enters a character string at the top of a line stack
Returns a pointer to the new stack
Paramters
  (1) stack -> current stack
  (2) lline -> the character string to be added
  (3) start -> the beginning index in the string line
  (3) end -> the final index in the string line
NOTE: A termination character '\0' will be added to the string
************************************************************************/
struct line_stacks *push_line(struct line_stacks *stack,char *lline,
                              int start, int end)
{
  struct line_stacks *root;
  int len;
  int i,j;
  
  root = (struct line_stacks *) malloc(sizeof(struct line_stacks));
  if (root == NULL) {
    fprintf(stderr,"%s\n","memory error");
    exit(1);
  }

  len = end - start;
  root->line = (char *) malloc((len+5)*sizeof(char));
  if (root->line == NULL) {
    fprintf(stderr,"%s\n","memory error");
    exit(1);
  }

  j = 0;
  for (i=start; i <= end; i++)
    root->line[j++] = lline[i];
  root->line[j] = '\0';

  root->next = stack;
  return(root);

}

/************************************************************************
POP_LINE
Removes the top structure from a line stack
Returns a pointer to the new stack
Paramters
  (1) stack -> current stack
************************************************************************/
struct line_stacks *pop_line(struct line_stacks *stack)
{
  struct line_stacks *root;

  if (stack == NULL) return(NULL);

  root = stack->next;
  free(stack->line);
  free(stack);
  return(root);
}

/************************************************************************
FLECS_ERROR
Prints out a flecs error message
Returns a pointer to the new stack
Paramters
  (1) type -> An integer indicating the type of error.
              0 : errors which which cause a line to be ignored.
	      1 : errors where a flecs structure has not been
	          terminated but should have been. In this case
		  the structure is terminated and the next line
		  is processed in the STANDARD state.
	      2 : warnings of probably errors which a sub-structure,
	          like and instance or ELSE statement, is encountered
	          without a parent structure. In this case the
		  line is echoed to the fortran file as is.
	      3 : errors where flecs structure is not terminated
	          when an END statement is encountered.
  (2) msg -> The particular error message
  (3) lline_num -> The line number of the error
************************************************************************/
void flecs_error(int type,char *msg,int lline_num)
{
  /* External Variables */
  extern int errnum;        /* Total number of flecs errors in a
			       subprogram */
  extern char *flecs_name;  /* name of file with flecs source code */

  /* Increment the error counter */
  errnum++;

  /* Print the error message */
  fprintf(stderr,"\"%s\", line %d:",flecs_name,lline_num);
  if (type == 0) fprintf(stderr," (I) ");
  else if (type == 1) fprintf(stderr," (T) ");
  else if (type == 2) fprintf(stderr," (E) ");
  fprintf(stderr,"%s\n",msg);
}

/************************************************************************
QUIT
Frees memory and closes files for flecs90 function
************************************************************************/
void quit(void)
{
  /* External Variables */
  extern FILE *flecs;       /* file pointer to flecs input file */
  extern FILE *fortran;     /* file pointer to fortran output file */
  extern char *flecs_name;  /* name of file with flecs source code */
  extern char *fortran_name;/* name of file with fortran code */
  extern char *line;        /* current line from flecs source file */
  extern char *error_msg;   /* error message string */
  extern char *flecs_proc;  /* name of in-line flecs procedure calls */
  extern char *scratch;     /* length 80 string for temporary use */

  free(flecs_name);
  free(fortran_name);
  free(line);
  free(scratch);
  free(error_msg);
  free(flecs_proc);
  fclose(flecs);
  fclose(fortran);
}
